@@ -1287,6 +1287,12 @@ impl Concat {
         let extra = if is_stack { 1 } else { 0 };
         Ok(check_canon_axes(nodes[0].info().rank() + extra, &[axis])?[0])
     }
+
+    pub fn new_signed_axis(nodes: Vec<CircuitRc>, axis: i64, name: Option<Name>) -> Result<Concat> {
+        let axis = Self::convert_axis(&nodes, axis, false)?;
+        let out = Self::try_new(nodes, axis, name)?;
+        Ok(out)
+    }
 }
 
 circuit_node_extra_impl!(Concat, self_hash_default);
@@ -1376,9 +1382,7 @@ impl Concat {
         axis: i64,
         name: Option<Name>,
     ) -> Result<PyClassInitializer<Concat>> {
-        let axis = Self::convert_axis(&nodes, axis, false)?;
-        let out = Self::try_new(nodes, axis, name)?;
-        Ok(out.into_init())
+        Self::new_signed_axis(nodes, axis, name).map(|x| x.into_init())
     }
 
     #[getter]
