@@ -22,6 +22,7 @@ use macro_rules_attribute::apply;
 pub use module::{Module, ModuleArgSpec, ModuleSpec};
 use named_axes::set_named_axes;
 use num_bigint::BigUint;
+use once_cell::sync::Lazy;
 use opaque_iterative_matcher::OpaqueIterativeMatcherVal;
 use print_html::{PrintHtmlOptions, PrintOptionsBase};
 use pyo3::{
@@ -32,7 +33,9 @@ use pyo3::{
 };
 use rr_util::{
     cached_lambda,
+    compact_data::U8Set,
     eq_by_big_hash::EqByBigHash,
+    name::Name,
     py_types::{
         reduction_to_ints, use_rust_comp, PyCallable, PyOpAtAxes, PyShape, Tensor, PY_CIRCUIT_ITEMS,
     },
@@ -119,8 +122,6 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     // we could enforce this sort of stuff with some proc macros, but seems like overkill atm.
 
     fn info(&self) -> &CachedCircuitInfo;
-    fn name(&self) -> Option<&str>;
-
     /// child -> child dim -> parent dim it's the same as
     /// same as = index on that dim commutes w/ self
     /// must be None if not or if unsure
@@ -165,7 +166,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         *uuid::Uuid::nil().as_bytes()
     }
 
-    fn get_autoname(&self) -> Result<Option<String>>;
+    fn get_autoname(&self) -> Result<Option<Name>>;
 
     // ==== default only section ===
     // FUNCTIONS BELOW HERE *shouldn't* be overridden by implementors!
@@ -189,7 +190,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     {
         self.initial_init_info_impl()
     }
-    fn rename(self, new_name: Option<String>) -> Self
+    fn rename(self, new_name: Option<Name>) -> Self
     where
         Self: Sized,
     {
@@ -214,9 +215,6 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     {
         self.clone().rc()
     }
-    fn name_cloned(&self) -> Option<String> {
-        self.name().map(|x| x.to_owned())
-    }
 
     // TODO: add more map utils for non_free as needed
 
@@ -360,7 +358,6 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     fn get_hash(&self) -> HashBytes {
         self.info().hash
     }
-
     fn rank(&self) -> usize {
         self.info().rank()
     }
@@ -374,7 +371,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         self.info().numel_usize()
     }
 
-    fn sum(&self, axes: &[i64], name: Option<String>) -> Result<Einsum>
+    fn sum(&self, axes: &[i64], name: Option<Name>) -> Result<Einsum>
     where
         Self: Clone,
     {
@@ -388,12 +385,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         ))
     }
 
-    fn mean(
-        &self,
-        axes: &[i64],
-        name: Option<String>,
-        scalar_name: Option<String>,
-    ) -> Result<Einsum>
+    fn mean(&self, axes: &[i64], name: Option<Name>, scalar_name: Option<Name>) -> Result<Einsum>
     where
         Self: Clone,
     {
@@ -409,7 +401,8 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
                     Scalar::nrc(
                         1. / (total_size as f64),
                         sv![],
-                        scalar_name.or_else(|| self.name().map(|s| format!("mean_div_{}", s))),
+                        scalar_name
+                            .or_else(|| self.info().name.map(|s| format!("mean_div_{}", s).into())),
                     ),
                     tu8v![],
                 )))
@@ -419,11 +412,11 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         ))
     }
 
-    fn reduce(&self, op_name: String, axes: &[i64], name: Option<String>) -> Result<Circuit>
+    fn reduce(&self, op_name: Name, axes: &[i64], name: Option<Name>) -> Result<Circuit>
     where
         Self: Clone,
     {
-        match op_name.as_str() {
+        match op_name.str() {
             "mean" => return self.mean(axes, name, None).map(CircuitNode::c),
             "sum" => return self.sum(axes, name).map(CircuitNode::c),
             _ => (),
@@ -443,35 +436,35 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         .unwrap()
         .c())
     }
-    fn min_(&self, axes: &[i64], name: Option<String>) -> Result<Circuit>
+    fn min_(&self, axes: &[i64], name: Option<Name>) -> Result<Circuit>
     where
         Self: Clone,
     {
-        self.reduce("min".to_owned(), axes, name)
+        self.reduce("min".into(), axes, name)
     }
-    fn max_(&self, axes: &[i64], name: Option<String>) -> Result<Circuit>
+    fn max_(&self, axes: &[i64], name: Option<Name>) -> Result<Circuit>
     where
         Self: Clone,
     {
-        self.reduce("max".to_owned(), axes, name)
+        self.reduce("max".into(), axes, name)
     }
-    fn add(&self, other: CircuitRc, name: Option<String>) -> Result<Add>
+    fn add(&self, other: CircuitRc, name: Option<Name>) -> Result<Add>
     where
         Self: Clone,
     {
         Add::try_new(vec![self.crc(), other], name)
     }
-    fn sub(&self, other: CircuitRc, name: Option<String>) -> Result<Add>
+    fn sub(&self, other: CircuitRc, name: Option<Name>) -> Result<Add>
     where
         Self: Clone,
     {
-        let other_neg_name = other.name().map(|s| format!("neg {}", s));
+        let other_neg_name = other.info().name.map(|s| format!("neg {}", s).into());
         self.add(
-            Einsum::scalar_mul(other, -1.0, other_neg_name, Some("neg_1".to_owned())).rc(),
+            Einsum::scalar_mul(other, -1.0, other_neg_name, Some("neg_1".into())).rc(),
             name,
         )
     }
-    fn mul(&self, other: CircuitRc, name: Option<String>) -> Result<Einsum>
+    fn mul(&self, other: CircuitRc, name: Option<Name>) -> Result<Einsum>
     where
         Self: Clone,
     {
@@ -480,15 +473,15 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     fn mul_scalar(
         &self,
         scalar: f64,
-        name: Option<String>,
-        scalar_name: Option<String>,
+        name: Option<Name>,
+        scalar_name: Option<Name>,
     ) -> Result<Einsum>
     where
         Self: Clone,
     {
         self.mul(Scalar::nrc(scalar, sv![], scalar_name), name)
     }
-    fn index(&self, index: TensorIndex, name: Option<String>) -> Result<Index>
+    fn index(&self, index: TensorIndex, name: Option<Name>) -> Result<Index>
     where
         Self: Clone,
     {
@@ -498,7 +491,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         &self,
         axes: Vec<usize>,
         counts: Vec<usize>,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Rearrange>
     where
         Self: Clone,
@@ -509,7 +502,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
             name,
         ))
     }
-    fn unsqueeze(&self, axes: Vec<usize>, name: Option<String>) -> Result<Rearrange>
+    fn unsqueeze(&self, axes: Vec<usize>, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
@@ -519,7 +512,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
             name,
         ))
     }
-    fn squeeze(&self, axes: PyOpAtAxes, name: Option<String>) -> Result<Rearrange>
+    fn squeeze(&self, axes: PyOpAtAxes, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
@@ -555,7 +548,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
             name,
         )
     }
-    fn flatten(&self, name: Option<String>) -> Result<Rearrange>
+    fn flatten(&self, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
@@ -565,13 +558,13 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
             name,
         ))
     }
-    fn unflatten(&self, shape: Shape, name: Option<String>) -> Result<Rearrange>
+    fn unflatten(&self, shape: Shape, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
         Rearrange::unflatten(self.crc(), shape, name)
     }
-    fn unflatten_axis(&self, axis: i64, shape: Shape, name: Option<String>) -> Result<Rearrange>
+    fn unflatten_axis(&self, axis: i64, shape: Shape, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
@@ -585,7 +578,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         &self,
         axis: usize,
         shape: Shape,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Rearrange>
     where
         Self: Clone,
@@ -596,13 +589,13 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
             name,
         ))
     }
-    fn rearrange(&self, spec: RearrangeSpec, name: Option<String>) -> Result<Rearrange>
+    fn rearrange(&self, spec: RearrangeSpec, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
         Rearrange::try_new(self.crc(), spec, name)
     }
-    fn rearrange_str(&self, string: &str, name: Option<String>) -> Result<Rearrange>
+    fn rearrange_str(&self, string: &str, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
@@ -614,7 +607,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     {
         match suffix {
             Some(suffix) => {
-                let name = self.name().map(|x| format!("{}_{}", x, suffix));
+                let name = self.info().name.map(|x| format!("{}_{}", x, suffix).into());
                 self.rename_impl(name)
             }
             None => self,
@@ -628,7 +621,7 @@ pub trait CircuitNodeAutoName: CircuitNode {
     /// - functions, like GeneralFunction and Cumulant, which use the standard function notation f(x1, ... xn)
     /// - postfix, like Rearrange and index, which behave like the unary operator "!". Parenthesis are added if needed
     /// utils to add the parenthesis can be found in circuit_utils.rs
-    fn auto_name(&self) -> Option<String>;
+    fn auto_name(&self) -> Option<Name>;
 
     const PRIORITY: OperatorPriority = OperatorPriority::NotOperator {};
     const CHILD_NAME_MAX_LEN: usize = 100; // TODO: fix me!
@@ -672,7 +665,7 @@ macro_rules! circuit_node_eq_ord_debug {
             fn cmp(&self, other: &Self) -> ::std::cmp::Ordering {
                 use $crate::prelude::*;
                 // name and then
-                (self.name(), self.info().hash).cmp(&(other.name(), other.info().hash))
+                (self.info().name, self.info().hash).cmp(&(other.info().name, other.info().hash))
             }
         }
 
@@ -882,7 +875,7 @@ macro_rules! define_circuit_union_impl {
                     (node, construct) => Ok(construct(node.init_info_impl(is_initial)?)))
             }
 
-            fn rename_impl(self, new_name: Option<String>) -> Self {
+            fn rename_impl(self, new_name: Option<Name>) -> Self {
                 [<$name:snake _map_construct>]!(self;
                     (node, construct) => construct(node.rename(new_name)))
             }
@@ -1119,9 +1112,6 @@ macro_rules! circuit_node_extra_impl {
             fn info_mut(&mut self) -> &mut $crate::CachedCircuitInfo {
                 &mut self.info
             }
-            fn name_mut(&mut self) -> &mut Option<String> {
-                &mut self.name
-            }
         }
         impl rr_util::eq_by_big_hash::EqByBigHash for $type_name {
             #[inline]
@@ -1153,7 +1143,7 @@ macro_rules! circuit_node_auto_impl {
     ($the_uuid:literal) => {
         $crate::circuit_node_auto_impl!($the_uuid, none);
 
-        fn get_autoname(&self) -> Result<Option<String>> {
+        fn get_autoname(&self) -> Result<Option<Name>> {
             Ok(self.auto_name())
         }
     };
@@ -1162,10 +1152,6 @@ macro_rules! circuit_node_auto_impl {
         fn info(&self) -> &$crate::CachedCircuitInfo {
             &self.info
         }
-        #[inline]
-        fn name(&self) -> Option<&str> {
-            self.name.as_deref()
-        }
         fn node_type_uuid(&self) -> [u8; 16] {
             Self::static_node_type_uuid()
         }
@@ -1186,7 +1172,7 @@ macro_rules! circuit_node_auto_impl {
     ($the_uuid:literal, no_autoname) => {
         $crate::circuit_node_auto_impl!($the_uuid, none);
 
-        fn get_autoname(&self) -> anyhow::Result<Option<String>> {
+        fn get_autoname(&self) -> anyhow::Result<Option<Name>> {
             anyhow::bail!("this node type doesn't support autoname");
         }
     };
@@ -1204,10 +1190,6 @@ impl<T: CircuitNodeDefer> CircuitNode for T {
         self.as_trait_obj().info()
     }
 
-    fn name(&self) -> Option<&str> {
-        self.as_trait_obj().name()
-    }
-
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
         self.as_trait_obj().child_axis_map()
     }
@@ -1254,7 +1236,7 @@ impl<T: CircuitNodeDefer> CircuitNode for T {
         self.custom_rc()
     }
 
-    fn get_autoname(&self) -> Result<Option<String>> {
+    fn get_autoname(&self) -> Result<Option<Name>> {
         self.as_trait_obj().get_autoname()
     }
 }
@@ -1267,7 +1249,8 @@ pub struct CachedCircuitInfo {
     pub device_dtype: TorchDeviceDtypeOp,
     pub named_axes: NamedAxes,
     pub symbolic_size_constraints: SymbolicSizeConstraints,
-    pub free_symbols: IndexSet<Symbol>,
+    pub free_symbols: Option<Arc<IndexSet<Symbol>>>,
+    pub name: Option<Name>,
 }
 
 #[derive(Clone, Copy, Hash, Debug, Eq, PartialEq)]
@@ -1329,10 +1312,16 @@ impl Not for CircuitFlags {
     }
 }
 
-/// don't want to print hash with Debug, for now just print shape
+/// don't want to print hash with Debug; print selected fields
 impl Debug for CachedCircuitInfo {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(f, "{:?}", self.shape)
+        write!(
+            f,
+            "{} {:?} {:?}",
+            self.name.map_or("", |n| n.str()),
+            self.shape,
+            self.device_dtype,
+        )
     }
 }
 
@@ -1386,6 +1375,15 @@ impl CachedCircuitInfo {
     pub fn use_autoname(&self) -> bool {
         self.flags.check(CircuitFlags::USE_AUTONAME)
     }
+
+    pub fn get_raw_free_symbols(&self) -> &IndexSet<Symbol> {
+        static NULL_INDEX_SET: Lazy<IndexSet<Symbol>> = Lazy::new(|| IndexSet::default());
+        if let Some(z) = &self.free_symbols {
+            &**z
+        } else {
+            &*NULL_INDEX_SET
+        }
+    }
 }
 
 #[apply(python_error_exception)]
@@ -1451,14 +1449,14 @@ pub enum ConstructError {
     EinsumWrongNumChildren {
         actual_num_children: usize,
         expected_num_children_based_on_axes: usize,
-        einsum_name: Option<String>,
+        einsum_name: Option<Name>,
     },
 
     #[error("len_axes={len_axes} != child_len_shape={child_len_shape}\neinsum_name={einsum_name:?} child_circuit={child_circuit:?} ({e_name})")]
     EinsumLenShapeDifferentFromAxes {
         len_axes: usize,
         child_len_shape: usize,
-        einsum_name: Option<String>,
+        einsum_name: Option<Name>,
         child_circuit: CircuitRc,
     },
     #[error("new_size={new_size}!=existing_size={existing_size} for axis={axis}\neinsum_name={einsum_name:?}, size introduced with child_circuit={child_circuit:?} ({e_name})")]
@@ -1466,13 +1464,13 @@ pub enum ConstructError {
         new_size: SymbolicSizeProduct,
         existing_size: SymbolicSizeProduct,
         axis: usize,
-        einsum_name: Option<String>,
+        einsum_name: Option<Name>,
         child_circuit: CircuitRc,
     },
     #[error("circuit_name={circuit_name:?} all_input_axes={all_input_axes:?} output_axes={output_axes:?} ({e_name})")]
     EinsumOutputNotSubset {
-        circuit_name: Option<String>,
-        all_input_axes: HashSet<u8>,
+        circuit_name: Option<Name>,
+        all_input_axes: U8Set,
         output_axes: EinsumAxes,
     },
 
@@ -1508,7 +1506,7 @@ pub enum ConstructError {
     ScatterIndexTypeUnimplemented { index: TensorIndex },
 
     #[error("Unknown GeneralFunction name {spec_name} ({e_name})")]
-    UnknownGeneralFunction { spec_name: String },
+    UnknownGeneralFunction { spec_name: Name },
 
     #[error("Module wrong number of children, expected {expected} got {got}\narg_specs={arg_specs:?} nodes={nodes:?} ({e_name})")]
     ModuleWrongNumberChildren {
@@ -1522,12 +1520,12 @@ pub enum ConstructError {
         "Module got unknown keyword argument, {argument}, all_module_inputs={all_module_inputs:?} ({e_name})"
     )]
     ModuleUnknownArgument {
-        argument: String,
-        all_module_inputs: Vec<String>,
+        argument: Name,
+        all_module_inputs: Vec<Name>,
     },
 
     #[error("missing_arguments={missing_arguments:?} ({e_name})")]
-    ModuleMissingNames { missing_arguments: Vec<String> },
+    ModuleMissingNames { missing_arguments: Vec<Name> },
 
     #[error("Module extract: not all leaves present in circuit, {subcirc:?} ({e_name})")]
     ModuleExtractNotPresent { subcirc: CircuitRc },
@@ -1536,7 +1534,7 @@ pub enum ConstructError {
     ModuleSomeArgsNamedNone { symbols_named_none: Vec<Symbol> },
 
     #[error("dup_names={dup_names:?} ({e_name})")]
-    ModuleArgsDupNames { dup_names: HashMap<String, usize> },
+    ModuleArgsDupNames { dup_names: HashMap<Name, usize> },
 
     #[error("spec_circuit={spec_circuit:?} missing_symbols={missing_symbols:?} ({e_name})")]
     ModuleSomeArgsNotPresent {
@@ -1614,7 +1612,7 @@ impl fmt::Display for CircuitRc {
 
 pub fn make_children_zero<T: CircuitNode>(circuit: &T) -> T {
     circuit.map_children_unwrap(&mut |child: CircuitRc| {
-        Scalar::new(0.0, child.info().shape.clone(), child.name_cloned()).rc()
+        Scalar::new(0.0, child.info().shape.clone(), child.info().name).rc()
     })
 }
 
@@ -2016,7 +2014,7 @@ impl CircuitNodeInit for CircuitRc {
         Ok(self.c().clone().init_info_impl(is_initial)?.rc())
     }
 
-    fn rename_impl(self, new_name: Option<String>) -> Self {
+    fn rename_impl(self, new_name: Option<Name>) -> Self {
         self.c().clone().rename(new_name).rc()
     }
 
@@ -2121,15 +2119,15 @@ impl PyCircuitBase {
 
     #[getter]
     fn name(&self) -> &str {
-        self.0.name().unwrap_or("")
+        Name::str_maybe_empty(self.0.info().name)
     }
 
     #[getter]
-    fn op_name(&self) -> Option<&str> {
-        self.0.name()
+    fn op_name(&self) -> Option<&'static str> {
+        self.0.info().name.map(|z| z.into())
     }
 
-    fn get_autoname(&self) -> Result<Option<String>> {
+    fn get_autoname(&self) -> Result<Option<Name>> {
         self.0.get_autoname()
     }
 
@@ -2138,10 +2136,12 @@ impl PyCircuitBase {
         self.info().symbolic_size_constraints.clone()
     }
 
+    #[getter]
     fn children(&self) -> Vec<CircuitRc> {
         self.0.children().collect()
     }
 
+    #[getter]
     fn non_free_children(&self) -> Vec<CircuitRc> {
         self.0.non_free_children().collect()
     }
@@ -2199,11 +2199,12 @@ impl PyCircuitBase {
     fn print_html(&self, options: PrintHtmlOptions) -> Result<()> {
         options.print(vec![(*self.0).crc()])
     }
-
+    #[getter]
     fn numel(&self) -> BigUint {
         self.0.info().numel()
     }
 
+    #[getter]
     fn rank(&self) -> usize {
         self.0.info().rank()
     }
@@ -2234,6 +2235,7 @@ impl PyCircuitBase {
         self.0.map_children(|child| pycall!(f, (child,), anyhow))
     }
 
+    #[getter]
     fn num_children(&self) -> usize {
         self.0.num_children()
     }
@@ -2262,7 +2264,7 @@ impl PyCircuitBase {
         get_compatible_dtype(&***self)
     }
 
-    fn rename(&self, name: Option<String>) -> CircuitRc {
+    fn rename(&self, name: Option<Name>) -> CircuitRc {
         self.0.clone().rename(name)
     }
 
@@ -2280,15 +2282,15 @@ impl PyCircuitBase {
 
     fn reduce(
         &self,
-        op_name: String,
+        op_name: Name,
         axis: Option<PyOpAtAxes>,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Circuit> {
         self.0
             .reduce(op_name, &reduction_to_ints(axis, self.info().rank()), name)
     }
 
-    fn sum(&self, axis: Option<PyOpAtAxes>, name: Option<String>) -> Result<Einsum> {
+    fn sum(&self, axis: Option<PyOpAtAxes>, name: Option<Name>) -> Result<Einsum> {
         Ok(self
             .0
             .sum(&reduction_to_ints(axis, self.info().rank()), name)?)
@@ -2296,8 +2298,8 @@ impl PyCircuitBase {
     fn mean(
         &self,
         axis: Option<PyOpAtAxes>,
-        name: Option<String>,
-        scalar_name: Option<String>,
+        name: Option<Name>,
+        scalar_name: Option<Name>,
     ) -> Result<Einsum> {
         Ok(self.0.mean(
             &reduction_to_ints(axis, self.info().rank()),
@@ -2305,44 +2307,44 @@ impl PyCircuitBase {
             scalar_name,
         )?)
     }
-    fn min(&self, axis: Option<PyOpAtAxes>, name: Option<String>) -> Result<CircuitRc> {
+    fn min(&self, axis: Option<PyOpAtAxes>, name: Option<Name>) -> Result<CircuitRc> {
         Ok(self
             .0
             .min_(&reduction_to_ints(axis, self.info().rank()), name)?
             .rc())
     }
-    fn max(&self, axis: Option<PyOpAtAxes>, name: Option<String>) -> Result<CircuitRc> {
+    fn max(&self, axis: Option<PyOpAtAxes>, name: Option<Name>) -> Result<CircuitRc> {
         Ok(self
             .0
             .max_(&reduction_to_ints(axis, self.info().rank()), name)?
             .rc())
     }
 
-    fn add(&self, other: CircuitRc, name: Option<String>) -> Result<Add> {
+    fn add(&self, other: CircuitRc, name: Option<Name>) -> Result<Add> {
         self.0.add(other, name)
     }
-    fn sub(&self, other: CircuitRc, name: Option<String>) -> Result<Add> {
+    fn sub(&self, other: CircuitRc, name: Option<Name>) -> Result<Add> {
         self.0.sub(other, name)
     }
-    fn mul(&self, other: CircuitRc, name: Option<String>) -> Result<Einsum> {
+    fn mul(&self, other: CircuitRc, name: Option<Name>) -> Result<Einsum> {
         self.0.mul(other, name)
     }
     fn mul_scalar(
         &self,
         scalar: f64,
-        name: Option<String>,
-        scalar_name: Option<String>,
+        name: Option<Name>,
+        scalar_name: Option<Name>,
     ) -> Result<Einsum> {
         self.0.mul_scalar(scalar, name, scalar_name)
     }
-    fn index(&self, index: TensorIndex, name: Option<String>) -> Result<Index> {
+    fn index(&self, index: TensorIndex, name: Option<Name>) -> Result<Index> {
         self.0.index(index, name)
     }
     fn expand_at_axes(
         &self,
         axes: PyOpAtAxes,
         counts: Option<PyCountsAtAxes>,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Rearrange> {
         Ok(Rearrange::new(
             self.0.crc(),
@@ -2350,25 +2352,25 @@ impl PyCircuitBase {
             name,
         ))
     }
-    fn unsqueeze(&self, axes: PyOpAtAxes, name: Option<String>) -> Result<Rearrange> {
+    fn unsqueeze(&self, axes: PyOpAtAxes, name: Option<Name>) -> Result<Rearrange> {
         self.expand_at_axes(axes, None, name)
     }
-    fn squeeze(&self, axes: PyOpAtAxes, name: Option<String>) -> Result<Rearrange> {
+    fn squeeze(&self, axes: PyOpAtAxes, name: Option<Name>) -> Result<Rearrange> {
         CircuitNode::squeeze(&self.clone().crc(), axes, name)
     }
-    fn flatten(&self, name: Option<String>) -> Result<Rearrange> {
+    fn flatten(&self, name: Option<Name>) -> Result<Rearrange> {
         self.0.flatten(name)
     }
-    fn unflatten(&self, shape: Shape, name: Option<String>) -> Result<Rearrange> {
+    fn unflatten(&self, shape: Shape, name: Option<Name>) -> Result<Rearrange> {
         self.0.unflatten(shape, name)
     }
-    fn unflatten_axis(&self, axis: i64, shape: Shape, name: Option<String>) -> Result<Rearrange> {
+    fn unflatten_axis(&self, axis: i64, shape: Shape, name: Option<Name>) -> Result<Rearrange> {
         self.0.unflatten_axis(axis, shape, name)
     }
-    fn rearrange(&self, spec: RearrangeSpec, name: Option<String>) -> Result<Rearrange> {
+    fn rearrange(&self, spec: RearrangeSpec, name: Option<Name>) -> Result<Rearrange> {
         self.0.rearrange(spec, name)
     }
-    fn rearrange_str(&self, string: &str, name: Option<String>) -> Result<Rearrange> {
+    fn rearrange_str(&self, string: &str, name: Option<Name>) -> Result<Rearrange> {
         self.0.rearrange_str(string, name)
     }
     pub fn is_irreducible_node(&self) -> bool {
@@ -2459,6 +2461,10 @@ impl PyCircuitBase {
     pub fn are_any_found(&self, matcher: OpaqueIterativeMatcherVal) -> PyResult<bool> {
         matcher.are_any_found(self.0.clone())
     }
+
+    pub fn symbolic_shape(&self) -> Vec<SymbolicSizeProduct> {
+        self.info().shape.iter().map(|&z| z.into()).collect()
+    }
 }
 
 pub fn get_compatible_dtype(circ: &Circuit) -> TorchDeviceDtype {
