@@ -13,7 +13,8 @@ use rr_util::{
     symbolic_size::{SymbolicSizeProduct, SIZE_PROD_MATCH},
     tensor_db::ensure_all_tensors_local,
     tensor_util::{
-        parse_numeric, ParseError, PyParseError, Shape, TensorIndex, TorchDeviceDtypeOp,
+        check_canon_idxs, parse_numeric, ParseError, PyParseError, Shape, TensorIndex,
+        TorchDeviceDtypeOp,
     },
     timed,
     util::{counts_g_1, flip_op_result, is_unique, NamedAxes},
@@ -836,7 +837,7 @@ impl Parser {
                 extra_should_be_empty()?;
                 Add::try_new(children, ps.name)?.rc()
             }
-            T::Concat => Concat::try_new(
+            T::Concat => Concat::new_signed_axis(
                 children,
                 parse_numeric(&ps.extra).context("failed to parse out concat axis")?,
                 ps.name,
