@@ -25,7 +25,7 @@ use rustc_hash::FxHashMap as HashMap;
 use crate::{
     canonicalize::deep_canonicalize,
     compiler_heuristics::deep_maybe_distribute,
-    compiler_strip::{remove_autotags, strip_names},
+    compiler_strip::strip_names_and_tags,
     deep_rewrite::{
         compiler_simp, deep_heuristic_nest_adds, deep_optimize_einsums, deep_push_down_index_raw,
     },
@@ -81,6 +81,8 @@ pub struct OptimizationSettings {
     pub optimization_parallelism: Option<usize>,
     #[pyo3(get, set)]
     pub device_dtype: Option<TorchDeviceDtype>,
+    #[pyo3(get, set)]
+    pub keep_all_names: bool,
 }
 
 impl Default for OptimizationSettings {
@@ -105,6 +107,7 @@ impl Default for OptimizationSettings {
             save_circ: false,
             optimization_parallelism: Some(8),
             device_dtype: None,
+            keep_all_names: false,
         }
     }
 }
@@ -131,7 +134,8 @@ impl OptimizationSettings {
         log_slow_einsums = "OptimizationSettings::default().log_slow_einsums",
         save_circ = "OptimizationSettings::default().save_circ",
         optimization_parallelism = "None",
-        device_dtype = "None"
+        device_dtype = "None",
+        keep_all_names = "OptimizationSettings::default().keep_all_names"
     )]
     fn new(
         verbose: usize,
@@ -153,6 +157,7 @@ impl OptimizationSettings {
         save_circ: bool,
         optimization_parallelism: Option<usize>,
         device_dtype: Option<TorchDeviceDtype>,
+        keep_all_names: bool,
     ) -> Self {
         Self {
             verbose,
@@ -174,6 +179,7 @@ impl OptimizationSettings {
             save_circ,
             optimization_parallelism,
             device_dtype,
+            keep_all_names,
         }
     }
 }
@@ -239,8 +245,11 @@ pub fn optimize_circuit(
     context: &mut OptimizationContext,
 ) -> Result<CircuitRc> {
     let print_timings = context.settings.verbose >= 2;
-    let circuit = timed!(strip_names(circuit, context, true), 10, print_timings);
-    let circuit = remove_autotags(circuit);
+    let circuit = if !context.settings.keep_all_names {
+        timed!(strip_names_and_tags(circuit, context), 10, print_timings)
+    } else {
+        circuit
+    };
     let circuit = conform_all_modules(circuit);
     let circuit = inline_single_callsite_modules(circuit);
     if is_python_running() && any_children_with_symbolic_sizes(circuit.clone()) {
@@ -388,7 +397,7 @@ pub fn optimize_and_evaluate_many(
     }
 
     let schedule = optimize_to_schedule_many(circuits, settings)?;
-    Ok(schedule.evaluate_many(settings))
+    schedule.evaluate_many(settings)
 }
 
 #[pyfunction(settings = "Default::default()")]
