@@ -1,5 +1,5 @@
 // temporary schedule sending setup
-use anyhow::Result;
+use anyhow::{Context, Result};
 use circuit_base::{parsing::Parser, print::PrintOptions, CircuitNode, IrreducibleNode, Scalar};
 use miniserde::{json, Deserialize, Serialize};
 use pyo3::{prelude::*, types::PyByteArray};
@@ -162,6 +162,11 @@ impl TryFrom<&Schedule> for ScheduleToSend {
     type Error = anyhow::Error;
 
     fn try_from(x: &Schedule) -> Result<Self, Self::Error> {
+        x.check_no_syms().context(concat!(
+            "schedules with symbols can't be sent\n",
+            "(TODO: we should maybe avoid this check so that you can ",
+            "always serialize schedules, but ScheduleToSend doesn't eror check well...)"
+        ))?;
         fn options() -> PrintOptions {
             PrintOptions {
                 shape_only_when_necessary: false,
