@@ -26,6 +26,7 @@ pub struct PyUtils {
     tensor_scale: PyObject,
     pub generalfunctions: std::collections::HashMap<String, PyObject>,
     pub gen_index_function: PyObject,
+    pub pow: PyObject,
     einsum: PyObject,
     make_diagonal: PyObject,
     pub print: PyObject,
@@ -81,6 +82,7 @@ pub static PY_UTILS: GILLazyPy<PyUtils> = GILLazyPy::new_py(|py| {
             .unwrap()
             .extract()
             .unwrap(),
+        pow: get("pow"),
         einsum: get("einsum"),
         make_diagonal: get("make_diagonal"),
         print: get_builtin("print"),
